#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Tests for `torchani_step` package."""

import pytest  # noqa: F401
import torchani_step  # noqa: F401


def test_construction():
    """Just create an object and test its type."""
    result = torchani_step.TorchANI()
    assert str(type(result)) == "<class 'torchani_step.torchani.TorchANI'>"
